// lexer.h cK\̎͊
#ifndef __MERCURY_REGEX_LEXER__
#define __MERCURY_REGEX_LEXER__

#include <utility>                      // operator!=, operator<=, operator>, operator>=
#include "traits.h"

namespace mercury
{
	namespace _regex
	{
		using namespace std::rel_ops;

		// g[N̎
		typedef enum tagTOKEN_TYPE
		{
			TOKEN_CHARACTER,                        // ʏ̕
			TOKEN_CLASS,                            // NX
			TOKEN_NUMBER,                           // l
			TOKEN_DISJUNCTION,                      // I
			TOKEN_REPEAT0,                          // 0ȏ̌JԂ
			TOKEN_REPEAT1,                          // 1ȏ̌JԂ
			TOKEN_REPEAT01,                         // 0܂1̌JԂ
			TOKEN_REPEAT_BEGIN,                     // ʉJn
			TOKEN_REPEAT_SEPARATOR,                 // ʉ؂
			TOKEN_REPEAT_END,                       // ʉI
			TOKEN_ANY,                              // Cӂ1
			TOKEN_SET_BEGIN,                        // WJn
			TOKEN_SET_RANGE,                        // ͈
			TOKEN_SET_NOT,                          // W
			TOKEN_SET_END,                          // WI
			TOKEN_GROUP_BEGIN,                      // O[vJn
			TOKEN_GROUP_END,                        // O[vI
			TOKEN_EOS,                              // I[
		} TOKEN_TYPE;


		// g[N\
		template<typename _Char>
		struct token
		{
			TOKEN_TYPE type;

			_Char     ch;                           // @@@itype=TOKEN_CHARACTERj
			int       num;                          // l@@@itype=TOKEN_NUMBERj
			CHARCLASS cls;                          // NXitype=TOKEN_CLASSj
			bool      invert;                       // NX𔽓]邩H
		};


		// ͊
		template<typename _Char, typename _InputIterator, typename _Traits = regex_traits<_Char> >
		class lexer
		{
		public:
			typedef token<_Char> token_t;

		public:
			// RXgN^iwj
			lexer(const _InputIterator &pattern_begin, const _InputIterator &pattern_end)
				: m_pattern_begin(pattern_begin), m_pattern_end(pattern_end), m_context(CTX_NORMAL), m_type_before(TOKEN_CHARACTER),
				  m_head(false), m_tail(false)
			{
				_init();
			}


			bool is_head(void) const { return m_head; }
			bool is_tail(void) const { return m_tail; }


			// ̃g[N擾
			token_t get_token(void)
			{
				// ̏I[
				if(m_now == m_pattern_end)
				{
					const token_t tk = {TOKEN_EOS};
					return tk;
				}

				const _InputIterator prev = m_now++;
				const _Char          ch   = *prev;
				token_t tk = {TOKEN_CHARACTER};
				tk.ch = ch;

				// GXP[v
				if(ch == _Traits::meta_escape())
				{
					if(m_now != m_pattern_end)
					{
						tk = _token_escape(*m_now++);
					}
				}
				else
				{
					switch(m_context)
					{
					case CTX_NORMAL:                        // ʏ
						tk = _token_normal(ch);
						break;

					case CTX_SET:                           // W
						tk = _token_set(ch, prev);
						break;

					case CTX_REPEAT:                        // ʉ
						tk = _token_repeat(ch, prev);
						break;
					}
				}
				m_type_before = tk.type;
				return tk;
			}

		private:
			_InputIterator m_pattern_begin;
			_InputIterator m_pattern_end;
			_InputIterator m_now;

			// i݈ʒuj
			typedef enum tagCONTEXT
			{
				CTX_NORMAL,                             // ʏ̕
				CTX_SET,                                // W
				CTX_REPEAT,                             // ʉ
			} CONTEXT;
			CONTEXT    m_context;
			TOKEN_TYPE m_type_before;

			bool m_head;
			bool m_tail;

			// 
			void _init(void)
			{
				m_now = m_pattern_begin;
				if(m_now != m_pattern_end && *m_now == _Traits::meta_head())
				{
					m_head = true;
					m_now++;
				}
			}

		private:
			// GXP[v̌̃g[N
			token_t _token_escape(const _Char ch)
			{
				token_t tk;

				// {block}
				// ꕶ
				{
					const struct
					{
						_Char ch;
						_Char ch2;
					}
					char_array[] =
					{
						{_Traits::escape_bel(), _Traits::cntrl_bel()},
						{_Traits::escape_ff (), _Traits::cntrl_ff ()},
						{_Traits::escape_lf (), _Traits::cntrl_lf ()},
						{_Traits::escape_cr (), _Traits::cntrl_cr ()},
						{_Traits::escape_ht (), _Traits::cntrl_ht ()},
						{_Traits::escape_vt (), _Traits::cntrl_vt ()},
						{_Traits::escape_esc(), _Traits::cntrl_esc()},
					};

					for(size_t i = 0; i < mercury_countof(char_array); i++)
					{
						if(ch == char_array[i].ch)
						{
							tk.type = TOKEN_CHARACTER;
							tk.ch   = char_array[i].ch2;
							goto exit;
						}
					}
				}

				// {block}
				// NX
				{
					const struct
					{
						_Char     ch;
						CHARCLASS cls;
						bool      invert;
					}
					class_array[] =
					{
						{_Traits::escape_digit       (), CHARCLASS_DIGIT, false},
						{_Traits::escape_digit_invert(), CHARCLASS_DIGIT,  true},
						{_Traits::escape_upper       (), CHARCLASS_UPPER, false},
						{_Traits::escape_upper_invert(), CHARCLASS_UPPER,  true},
						{_Traits::escape_lower       (), CHARCLASS_LOWER, false},
						{_Traits::escape_lower_invert(), CHARCLASS_LOWER,  true},
						{_Traits::escape_space       (), CHARCLASS_SPACE, false},
						{_Traits::escape_space_invert(), CHARCLASS_SPACE,  true},
						{_Traits::escape_csym        (), CHARCLASS_CSYM , false},
						{_Traits::escape_csym_invert (), CHARCLASS_CSYM ,  true},
					};

					for(size_t i = 0; i < mercury_countof(class_array); i++)
					{
						if(ch == class_array[i].ch)
						{
							tk.type   = TOKEN_CLASS;
							tk.cls    = class_array[i].cls;
							tk.invert = class_array[i].invert;
							goto exit;
						}
					}
				}

				// ǂɂqbgȂΒʏ̕Ƃ
				tk.ch   = ch;
				tk.type = TOKEN_CHARACTER;

exit:
				return tk;
			}

			// ʏ̃̕g[N
			token_t _token_normal(const _Char ch)
			{
				token_t tk;
				tk.type = TOKEN_CHARACTER;
				tk.ch   = ch;

				// v
				if(ch == _Traits::meta_tail())
				{
					if(m_now == m_pattern_end)
					{
						m_tail  = true;
						tk.type = TOKEN_EOS;
					}
				}
				// I
				if(ch == _Traits::meta_disjunction())
				{
					tk.type = TOKEN_DISJUNCTION;
				}
				// 0ȏ̌JԂ
				if(ch == _Traits::meta_repeat0())
				{
					tk.type = TOKEN_REPEAT0;
				}
				// 1ȏ̌JԂ
				if(ch == _Traits::meta_repeat1())
				{
					tk.type = TOKEN_REPEAT1;
				}
				// 0܂1̌JԂ
				if(ch == _Traits::meta_repeat01())
				{
					tk.type = TOKEN_REPEAT01;
				}
				// ʉJn
				if(ch == _Traits::meta_repeat_begin())
				{
					m_context = CTX_REPEAT;
					tk.type = TOKEN_REPEAT_BEGIN;
				}
				// Cӂ1
				if(ch == _Traits::meta_any())
				{
					tk.type = TOKEN_ANY;
				}
				// WJn
				if(ch == _Traits::meta_set_begin())
				{
					m_context = CTX_SET;
					tk.type = TOKEN_SET_BEGIN;
				}
				// O[vJn
				if(ch == _Traits::meta_group_begin())
				{
					tk.type = TOKEN_GROUP_BEGIN;
				}
				// O[vI
				if(ch == _Traits::meta_group_end())
				{
					tk.type = TOKEN_GROUP_END;
				}

				return tk;
			}

			// W̃g[N
			token_t _token_set(const _Char ch, const _InputIterator &prev)
			{
				token_t tk;
				tk.type = TOKEN_CHARACTER;
				tk.ch   = ch;

				// NX
				const CHARCLASS cls = _Traits::get_charclass(prev, m_pattern_end, m_now);
				if(cls != CHARCLASS_NONE)
				{
					tk.type   = TOKEN_CLASS;
					tk.cls    = cls;
					tk.invert = false;
				}
				// W
				if(ch == _Traits::meta_set_not())
				{
					if(m_type_before == TOKEN_SET_BEGIN)
					{
						tk.type = TOKEN_SET_NOT;
					}
				}
				// ͈
				if(ch == _Traits::meta_set_range())
				{
					// Õg[NŁAWIłȂΕ͈͂̃g[N
					if(m_type_before == TOKEN_CHARACTER && m_now != m_pattern_end && *m_now != _Traits::meta_set_end())
					{
						tk.type = TOKEN_SET_RANGE;
					}
				}
				// WI
				if(ch == _Traits::meta_set_end())
				{
					m_context = CTX_NORMAL;
					tk.type = TOKEN_SET_END;
				}

				return tk;
			}

			// ʉ̃g[N
			token_t _token_repeat(const _Char ch, const _InputIterator &prev)
			{
				token_t tk;
				tk.type = TOKEN_CHARACTER;
				tk.ch   = ch;

				// l
				tk.num = _Traits::number(prev, m_pattern_end, m_now);
				if(tk.num != -1)
				{
					tk.type = TOKEN_NUMBER;
				}
				// ʉ؂
				if(ch == _Traits::meta_repeat_separator())
				{
					tk.type = TOKEN_REPEAT_SEPARATOR;
				}
				// ʉI
				if(ch == _Traits::meta_repeat_end())
				{
					m_context = CTX_NORMAL;
					tk.type = TOKEN_REPEAT_END;
				}

				return tk;
			}
		};
	}
}

#endif // __MERCURY_REGEX_LEXER__
